/*
	File:		Node.cpp

	Contains:	Implementation of Node class, a node in an n-way tree

	Owned by:	Reggie Adkins

	Copyright:	 1993-1996 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <3>	 5/24/96	jpa		Fixed header comment.
		 <2>	 5/24/96	jpa		1.1MRD: Inlined many methods for
									smaller/faster code.

	To Do:
*/

#ifndef _NODE_
#include "Node.h"
#endif

#ifndef _EXCEPT_
#include "Except.h"
#endif

#ifndef _ODDEBUG_
#include "ODDebug.h"
#endif


#pragma segment ODNode

//======================================================================================
// Class Node
//======================================================================================

//------------------------------------------------------------------------------
// Node::Node
//
// Description
//------------------------------------------------------------------------------

Node::Node() 
{
	fParent = kODNULL; 
}

//------------------------------------------------------------------------------
// Node::~Node
//
// Description
//------------------------------------------------------------------------------

Node::~Node() 
{
}

//------------------------------------------------------------------------------
// Node::Size
//
// Description
//------------------------------------------------------------------------------

ODULong Node::Size() 
{
	ODULong size = 0;
	
	Node* node = this->FirstTopDown();
	while (node)
	{
		size++;
		node = node->NextTopDown(kODFrontToBack);
	}
	return size;
}

//------------------------------------------------------------------------------
// Node::GetNextSibling
//
// Description
//------------------------------------------------------------------------------

Node* Node::GetNextSibling()
{
	Node* parent = this->GetParent();
	return parent ? parent->GetChildAfter(this) : kODNULL;
}

//------------------------------------------------------------------------------
// Node::GetPreviousSibling
//
// Description
//------------------------------------------------------------------------------

Node* Node::GetPreviousSibling()
{
	Node* parent = this->GetParent();
	return parent ? parent->GetChildBefore(this) : kODNULL;
}

//------------------------------------------------------------------------------
// Node::NextTopDown
//
// Description
//------------------------------------------------------------------------------

Node* Node::NextTopDown(ODSiblingOrder siblingOrder)
{
	Node* child;
	if (siblingOrder == kODFrontToBack)
		child = this->GetFirstChild();
	else
		child = this->GetLastChild();
	
	if (child)
		return child;
	else 
	{
		Node* sibling;
		if (siblingOrder == kODFrontToBack)
			sibling = this->GetNextSibling();
		else 
			sibling = this->GetPreviousSibling();

		if (sibling) 
			return  sibling;
		else 
			return this->GetNextAunt(siblingOrder);
	}
	return kODNULL;
}

//------------------------------------------------------------------------------
// Node::GetNextAunt
//
// Description
//------------------------------------------------------------------------------

Node* Node::GetNextAunt(ODSiblingOrder siblingOrder)
{
	Node* parent = this->GetParent();
	
	if (parent)
	{
		Node* sibling;
		if (siblingOrder == kODFrontToBack)
			sibling = parent->GetNextSibling();
		else
			sibling = parent->GetPreviousSibling();
		
		if (sibling) 
			return sibling;
		else 
			return parent->GetNextAunt(siblingOrder);
	}
	else
		return kODNULL;
}

//------------------------------------------------------------------------------
// Node::FirstBottomUp
//
// Description
//------------------------------------------------------------------------------

Node* Node::FirstBottomUp(ODSiblingOrder siblingOrder)
{
	Node* child;
	if (siblingOrder == kODFrontToBack)
		child = this->GetFirstChild();
	else
		child = this->GetLastChild();

	if (child)
		return child->FirstBottomUp(siblingOrder);
	else
		return this;
}

//------------------------------------------------------------------------------
// Node::NextBottomUp
//
// Description
//------------------------------------------------------------------------------

Node* Node::NextBottomUp(ODSiblingOrder siblingOrder)
{
	Node* parent = this->GetParent();

	if (parent)
	{
		Node* sibling;
		if (siblingOrder == kODFrontToBack)
			sibling = parent->GetChildAfter(this);
		else
			sibling = parent->GetChildBefore(this);
		
		if (sibling)
			return sibling->FirstBottomUp(siblingOrder);
		else
			return parent;
	}
	else
		return kODNULL;	

	return kODNULL;
}

//======================================================================================
// NodeTraverser
//======================================================================================

//------------------------------------------------------------------------------
// NodeTraverser::NodeTraverser
//
// Description
//------------------------------------------------------------------------------

NodeTraverser::NodeTraverser(Node* root, 
					ODTraversalType traversalType, 
					ODSiblingOrder siblingOrder)
{
	fRoot = root;
	fCurrent = kODNULL;
	fTraversalType = traversalType;
	fSiblingOrder = siblingOrder;
}

//------------------------------------------------------------------------------
// NodeTraverser::First
//
// Description
//------------------------------------------------------------------------------

Node*	NodeTraverser::First()
{
	if (fRoot)
	{
		if (fTraversalType == kODTopDown)
			fCurrent = this->FirstTopDown(fRoot);
		else if (fTraversalType == kODBottomUp)
			fCurrent = this->FirstBottomUp(fRoot,fSiblingOrder);
		else if (fTraversalType == kODChildrenOnly)
		{
			if (fSiblingOrder == kODFrontToBack)
				fCurrent = fRoot->GetFirstChild();
			else
				fCurrent = fRoot->GetLastChild();
		}
		return fCurrent;
	}
	else
		return kODNULL;
}

//------------------------------------------------------------------------------
// NodeTraverser::Next
//
// Description
//------------------------------------------------------------------------------

Node*	NodeTraverser::Next()
{
	if (fCurrent)
	{
		if (fTraversalType == kODTopDown)
			fCurrent = this->NextTopDown(fCurrent, fSiblingOrder);
		else if (fTraversalType == kODBottomUp)
			fCurrent = this->NextBottomUp(fCurrent, fSiblingOrder);
		else if (fTraversalType == kODChildrenOnly)
		{
			if (fSiblingOrder == kODFrontToBack)
				fCurrent = fCurrent->GetNextSibling();
			else
				fCurrent = fCurrent->GetPreviousSibling();
		}
		return fCurrent;
	}
	else
		return kODNULL;
}

//------------------------------------------------------------------------------
// NodeTraverser::SkipChildren
//
// Description
//------------------------------------------------------------------------------

void NodeTraverser::SkipChildren()
{
	Node* next = kODNULL;
	Node* skipTo = kODNULL;
	Node* test = kODNULL;

	if (fCurrent)
	{
		if (fTraversalType == kODTopDown)
		{
			if (fSiblingOrder == kODFrontToBack)
				next = fCurrent->GetNextSibling();
			else
				next = fCurrent->GetPreviousSibling();

			if (!next)
				next = fCurrent->GetNextAunt(fSiblingOrder);

			test = fCurrent;
			while ( test && (test != next) )
			{
				skipTo = test;
				test = this->NextTopDown(test, fSiblingOrder);
			}

			fCurrent = skipTo;
		}
	}
}

//------------------------------------------------------------------------------
// NodeTraverser::NextTopDown
//
// Description
//------------------------------------------------------------------------------

Node* NodeTraverser::NextTopDown(Node* node, ODSiblingOrder siblingOrder)
{
	Node* child;
	if (siblingOrder == kODFrontToBack)
		child = node->GetFirstChild();
	else
		child = node->GetLastChild();
	
	if (child)
		return child;
	else if (node != fRoot)
	{
		Node* sibling;
		if (siblingOrder == kODFrontToBack)
			sibling = node->GetNextSibling();
		else 
			sibling = node->GetPreviousSibling();

		if (sibling) 
			return  sibling;
		else 
			return this->GetNextAunt(node, siblingOrder);
	}
	return kODNULL;
}

//------------------------------------------------------------------------------
// NodeTraverser::GetNextAunt
//
// Description
//------------------------------------------------------------------------------

Node* NodeTraverser::GetNextAunt(Node* node, ODSiblingOrder siblingOrder)
{
	Node* parent = node->GetParent();
	
	if (parent && (parent != fRoot))
	{
		Node* sibling;
		if (siblingOrder == kODFrontToBack)
			sibling = parent->GetNextSibling();
		else
			sibling = parent->GetPreviousSibling();
		
		if (sibling) 
			return sibling;
		else 
			return this->GetNextAunt(parent, siblingOrder);
	}
	else
		return kODNULL;
}

//------------------------------------------------------------------------------
// NodeTraverser::FirstBottomUp
//
// Description
//------------------------------------------------------------------------------

Node* NodeTraverser::FirstBottomUp(Node* node, ODSiblingOrder siblingOrder)
{
	Node* child;
	if (siblingOrder == kODFrontToBack)
		child = node->GetFirstChild();
	else
		child = node->GetLastChild();

	if (child)
		return this->FirstBottomUp(child, siblingOrder);
	else
		return node;
}

//------------------------------------------------------------------------------
// Node::Node
//
// Description
//------------------------------------------------------------------------------

Node* NodeTraverser::NextBottomUp(Node* node, ODSiblingOrder siblingOrder)
{
	if (node == fRoot)
		return kODNULL;
		
	Node* parent = node->GetParent();

	if (parent)
	{
		Node* sibling;
		if (siblingOrder == kODFrontToBack)
			sibling = parent->GetChildAfter(node);
		else
			sibling = parent->GetChildBefore(node);
		
		if (sibling)
			return this->FirstBottomUp(sibling, siblingOrder);
		else
			return parent;
	}
	else
		return kODNULL;	

	return kODNULL;
}

